LoadPackage("HAP");

######################################################################
# Run this code using: Read("support.g");
#
# This code defines a large number of useful functions. The following is a subset.
#
# AutoResolveNormalSeries(G)
#  inputs: G,      a group
#  outputs: R,     a group resolution of G
#
# GetDistinctRankTwoAbelianSubgroups(G)
#  inputs: G,      a group
#  outputs: uniquepairs, a list of rank-two abelian subgroups of G up to conjugation
#
# ObstructedQ(group,R,cc,g,h)
#  inputs: group,  a group
#          R,      the resolution of group
#          cc,     the vector description of a cocycle
#          g,h     the generators for a rank-two abelian subgroup
#
#  outputs: obs,   A boolean describing whether <g,h> is obstructed
#
# FindUnobstructed(group, R, cc)
#  inputs: group,  a group
#          R,      the resolution of group
#          cc,     the vector description of a cocycle
#
#  outputs: L,     A list of all unobstructed rank-two abelian subgroups
#                  up to conjugation
#
# SilentVectorToCocycle(vector,GR)
#  inputs: vector, a vector description of a cocycle, like those found in
#                  group construction
#  inputs: GR,     a group or the resolution of a group
#
#  outputs: c,     a 3-cocycle, which can be called as c(g1,g2,g3), which returns
#                  an integer mod the exponent of the resolution
#
# ResolutionExponent(R)
#  inputs: R,      a resolution
#
#  outputs: exp    the exponent of the resolution
#
# TotalMultiplierPhase(group, R, cc, g, h, gamma)
#  inputs: group, R, a group and its resolution
#          cc,     a cocycle vector
#          g,h,    a commuting pair
#          gamma,  a 2x2 matrix in SL_2(Z)
#  
#  outputs: phase, the multiplier phase induced on phi_{g,h} be applying the
#                  modular operator |_gamma
#
# ConstructShurCoveringHomomorphism(G,g)
#  Prints a character table of the Schur cover of the centralizer of g in G, which is a list of
#  all its projective representations.  Also prints the correspondence from conjugacy classes
#  in the cover to conjugacy classes in the centralizer of g.
#  
#  outputs: the homomorphism defined by the cover or false if it is trivial
#
# cg(cocycle,g,h1,h2,exp)
#  inputs: cocycle, a cocycle function (like from SilentVectorToCocycle)
#          g, h1, h2, group elements
#          exp,    the exponent of the resolution that defined cocycle
#
#  outputs: cg,    cg_g(h1,h2) as defined in the paper.
#                  Given as an integer mod exp.
#
######################################################################

getorder:= function(vector,exp)
    if ForAll(vector,x->x=0) then return 1; fi;
    return exp/Gcd(exp,Gcd(vector));
end;

EWithName:=function(classnames, classreps, name)
  local index;
  index:=1;
  while index<=Length(classnames) do
    if classnames[index]=name then return classreps[index]; fi;
    index:=index+1;
  od;
  return false;
end;

MoonshineEWithName:=function(G,name)
  return EWithName(G!.MoonshineClassNames,G!.MoonshineClassReps,name);
end;

# For a given G, picks an aribtrary normal series and tries to resolve based off that.
AutoResolveNormalSeries:=function(G)
  local normalseries, R, subgroup, normalsubgroups;
  normalseries:=[G];
  subgroup:=G;
  Print("AutoResolving normal series:");
  #normalsubgroups := NormalSubgroups(G);
  while Size(subgroup)>1 do
    Print(Size(subgroup),"...");
    normalsubgroups:=Filtered(NormalSubgroups(subgroup),n->IsNormal(G,n));
    normalsubgroups:=Filtered(normalsubgroups,n->Size(n)<Size(subgroup));
    subgroup:=normalsubgroups[1];
    Append(normalseries,[subgroup]);
    #normalsubgroups:=Filtered(normalsubgroups,n->IsSubgroup(subgroup,n));
  od;
  Print("1\n");
  return ResolutionNormalSeries(normalseries,4);
end;


IsOld:=function(center,p)
  local g,h,z1,z2;
  # p is a rank at most 2 subgroup
  if IsCyclic(p) then return true; fi;

  g:=p.1;h:=p.2;

  for z1 in center do
    for z2 in center do
      if IsCyclic(Group([z1*g,z2*h])) then return true; fi;
    od;
  od;
  return false;
end;


Reps:=function(list)
  return List(list,s->Representative(s));
end;

ConjugateToWhich:=function(group,sets,element)
    local filtered;
    filtered :=Filtered([1..Size(sets)],n->IsConjugate(group,element,sets[n]));
    if Size(filtered)=0 then return false; fi;
    return filtered[1];
end;

InWhich:=function(sets,element)
    local filtered;
    filtered :=Filtered([1..Size(sets)],n->element in sets[n]);
    if Size(filtered)=0 then return false; fi;
    return filtered[1];
end;

GetDistinctRankTwoAbelianSubgroups := function(group)
    local cc,pairs,greps,hreps,cent,centcc,g,h,uniquepairs,n,m,ns,debug;
    debug := 0;

    # Get all of the gh pairs
    pairs := [];

    cc := ConjugacyClasses(group);
    if debug>0 then Print("calculated conjugacy classes\n"); fi;

    greps := List(cc,c->Representative(c));


    for g in greps do
        cent := Centralizer(group,g);
        centcc := ConjugacyClasses(cent);
        hreps := List(centcc,c->Representative(c));
        for h in hreps do
            Add(pairs,Group(g,h));
        od;
    od;
    if debug>0 then Print("found pairs: there are ",Size(pairs)," pairs\n"); fi;

    # Get the list of unique pairs
    uniquepairs := [];
    ns := [1..Size(pairs)];
    for n in ns do
        Add(uniquepairs,pairs[n]);
        Unbind(ns[n]);
        for m in ns do
            if IsConjugate(group,pairs[m],pairs[n]) then
                Unbind(ns[m]);
            fi;
        od;
    od;
    if debug>0 then Print("isolated unique pairs\n"); fi;

    return uniquepairs;
end;

IsRankTwoSubgroupCongruentToSubgroupOfSubgroup := function(group, ranktwo, sylowsubgroup)
    local sranktwos,sranktwo;

    sranktwos := GetDistinctRankTwoAbelianSubgroups(sylowsubgroup);
    for sranktwo in sranktwos do
        if IsConjugate(group,ranktwo,sranktwos) then
            return true;
        fi;
    od;
    return false;
end;


# Usage: Resolution(G)
# Returns a resolution of G of length 4 in characteristic 0
Resolution:=G->ResolutionFiniteGroup(G,4);

ResolutionExponent:=R->Lcm(Homology(TensorWithIntegers(R),3));

# Usage: ListStandardCocycles(G) or ListStandardCocycles(R)
# Returns a list of standard cohomological generators for the 3rd
# group cohomology of G into Z/expZ, which is in U(1), where exp
# is the exponent of the cohomology
ListStandardVectors := function(GR)
    local R,TT,exp,M,slns,BM,cycles,cobou,len,com,basis,cohom,elementorders,ranges,allvectors,allcocycles,hom;

    if IsGroup(GR) then
        R:=Resolution(GR);
    else R:=GR; fi;


    TT:=TensorWithIntegers(R);
    hom:=Homology(TT,3);
    exp:=Lcm(hom);;
    M:=CocycleCondition(R,3);;
    slns:=BasisNullspaceModN(TransposedMat(M),exp);;
    BM:=BoundaryMatrix(TT,3);;
    cycles:=NullspaceIntMat(TransposedMat(BM));;
    cobou:=NullspaceIntMat(TransposedMat(cycles));;

    # If cobou is trivial, the rest of this fails, so we skip it.
    if not cobou=[] then
        len:=RankMat(cobou)+RankMat(cycles);
        com:=ComplementIntMat(IdentityMat(len),cobou);;
        basis:=com.complement;;
        cohom:=BaseIntersectionIntMats(basis,slns);;
        cohom:=cohom mod exp;;
    else
        cohom:= slns;
    fi;

    # At this point we have a basis.  I actually want all of the cohomology functions, so I need to sum over the basis
    elementorders := List(cohom,c->[0..getorder(c,exp)-1]);
    ranges := Cartesian(elementorders);
    #Print("ranges: ",ranges,"\n");
    allvectors := ranges*cohom mod exp;
    allvectors := DuplicateFreeList(allvectors);
    return allvectors;
end;

ListStandardCocycles:=function(GR)
    local R,allvectors,allcocycles,exp;
    if IsGroup(GR) then
        R:=Resolution(GR);
    else R:=GR; fi;

    exp:=Lcm(Homology(TensorWithIntegers(R),3));
    #if homlist=[] then exp:=1; else exp:=Lcm(Homology(TensorWithIntegers(R),3)); fi;
    #exp:=Lcm(Homology(TensorWithIntegers(R),3));
    allvectors:=ListStandardVectors(R);
    allcocycles := List(allvectors,sol->StandardCocycle(R,sol,3,exp));
    return allcocycles;
end;

# Turns a solution of the cocycle condition into the corresponding standard cocycle
VectorToCocycle:= function(vector,GR)
    local R,TT,hom,exp;

    if IsGroup(GR) then
        R:=Resolution(GR);
    else R:=GR; fi;

    TT:=TensorWithIntegers(R);
    hom:=Homology(TT,3);
    Print("Homology:", hom,"\n");
    exp:=Lcm(hom);;
    return StandardCocycle(R,vector,3,exp);
end;

SilentVectorToCocycle:= function(vector,GR)
    local R,TT,hom,exp;

    if IsGroup(GR) then
        R:=Resolution(GR);
    else R:=GR; fi;

    TT:=TensorWithIntegers(R);
    hom:=Homology(TT,3);
    exp:=Lcm(hom);;
    return StandardCocycle(R,vector,3,exp);
end;

groupexponent:=function(group)
    return Lcm(GroupHomology(group,3));
end;


# Code for the calculation of multiplier systems

cg:=function(cocycle,g,h1,h2,exp)
    return cocycle(g,h1,h2)+cocycle(h1,h2,(h1*h2)^-1*g*h1*h2)-cocycle(h1,h1^-1*g*h1,h2) mod exp;
end;

# Evaluation of kappa, from equation (54) of 1211.7074v2
kappa:= function(cocycle,g,a,b,exp)
    local list;
    list := List([1..Order(g)],n->cocycle(g^a,g^(b+a*n),g^a));
    return Sum(list) mod exp;
end;

# Evaluation of xi, from equation (50)
xi := function(cocycle,g,h,k,exp)
    return cg(cocycle,g,h,k,exp)-cg(cocycle,g,k,h^k,exp) mod exp;
end;

matchtargetkappa:=function(cocycle,exp,hreps,target)
    local i,h,ka;

    for i in [1..Size(hreps)] do
        h:=hreps[i];
        if IsBound(target[i]) then
            ka := kappa(cocycle,h,1,0,exp);
            if ka <> target[i] then
                return false;
            fi;
        fi;
    od;
    return true;
end;

# checks whether two pairs of elements are pairwise conjugate.
IsPairwiseConjugate:= function(group,g1,h1,g2,h2)
    local gk,hk,gkcoset,hkcoset;

    if not IsConjugate(group,g1,g2) or not IsConjugate(group,h1,h2) then
        return false;
    fi;

    gk := RepresentativeAction(group,g1,g2);
    gkcoset := RightCoset(Centralizer(group,g1),gk);

    hk := RepresentativeAction(group,h1,h2);
    hkcoset := RightCoset(Centralizer(group,h1),hk);

    return Size(Intersection(hkcoset,gkcoset))>0;
end;

# returns k such that k^-1gk=g1^k=g2 and h1^k=h2
PairwiseConjugateRepresentative:=function(group,g1,h1,g2,h2)
    local gk,hk,gkcoset,hkcoset;
    gk := RepresentativeAction(group,g1,g2);
    gkcoset := RightCoset(Centralizer(group,g1),gk);

    hk := RepresentativeAction(group,h1,h2);
    hkcoset := RightCoset(Centralizer(group,h1),hk);

    return Intersection(hkcoset,gkcoset)[1];
end;

######################################################################
# This code is a tool to find obstructions for a given group
# It ignore the cyclic sbugroups, since those are already known,
# and returns the non-cyclic unobstructed rank 2 abelian subgroups
######################################################################

ObstructedFromCocycleQ:=function(group,c,exp,g,h) #c is cocycle
    local pairs, p,rpairs, obs,gk,gkcoset,hk,hkcoset,ks,k,out,phase,stab;
    p:=Group(g,h);

    if not g*h = h*g then
        Print("Error: g and h do not commute.");
        return false;
    fi;

    obs := false; #obstruction found

    #### OBSTRUCTION 1
    if IsConjugate(group,g,g^-1) and IsConjugate(group,h,h^-1) then
        # determine by what they are conjugate
        gk := RepresentativeAction(group,g,g^-1);
        gkcoset := RightCoset(Centralizer(group,g),gk);

        hk := RepresentativeAction(group,h,h^-1);
        hkcoset := RightCoset(Centralizer(group,h),hk);

        ks := Intersection(gkcoset,hkcoset);
        #Print("Possible type 2 obstruction. Testing ",Size(ks)," elements.\n");

        for k in ks do
            phase:= (-cg(c,h,g,g^-1,exp)-cg(c,g^-1,h,h^-1,exp) + xi(c,g^-1,h^-1,k,exp)) mod exp;
            if phase <> 0 then
                obs := true;
                Print("Type 2 obstruction found from k=",k,"\n");
                break;
            fi;
        od;
    fi;

    #### OBSTRUCTION 2
    # We are going to take the ugly route by bruteforce checking all the possibilities.
    # I am pretty sure it doesn't matter what generators we take, but I might want to verify this later

    if not obs then
        stab := Centralizer(group,p);
        if Size(stab)>1 then
            #Print("Possible type 1 obstruction. Testing ",Size(stab)," elements.\n");
        fi;
        for k in stab do
            phase := xi(c,g,h,k,exp) mod exp;
            if phase <> 0 then
                obs := true;
                Print("Type 1 obstruction found from k=",k,"\n");
                break;
            fi;
        od;

        if obs then
        else
            #Print("Not obstructed\n");
        fi;
    fi;

    return obs;
end;

ObstructedQ:=function(group,R,cc,g,h)
  local c,exp;
  c := SilentVectorToCocycle(cc,R);
  exp:=Lcm(Homology(TensorWithIntegers(R),3));
  return ObstructedFromCocycleQ(group,c,exp,g,h);
end;

FindUnobstructed:=function(group, R, cc)
    local c,exp,pairs, p,rpairs, g,h,obs,gk,gkcoset,hk,hkcoset,ks,k,out,phase,stab;
    c := VectorToCocycle(cc,R);
    exp:=Lcm(Homology(TensorWithIntegers(R),3));

    ## Start searching for obstructions
    pairs:=GetDistinctRankTwoAbelianSubgroups(group);
    rpairs:=Filtered(pairs,p->not IsCyclic(p));  # We know there are no obstructions for the rank 1 subgroups
    out := [];

    for p in rpairs do
        Print("\n",p,"\n");
        g := p.2;
        h := p.1;
        obs := false; #obstruction found

        #### OBSTRUCTION 1
        if IsConjugate(group,g,g^-1) and IsConjugate(group,h,h^-1) then
            # determine by what they are conjugate
            gk := RepresentativeAction(group,g,g^-1);
            gkcoset := RightCoset(Centralizer(group,g),gk);

            hk := RepresentativeAction(group,h,h^-1);
            hkcoset := RightCoset(Centralizer(group,h),hk);

            ks := Intersection(gkcoset,hkcoset);
            Print("Possible type 2 obstruction. Testing ",Size(ks)," elements.\n");

            for k in ks do
                phase:= (-cg(c,h,g,g^-1,exp)-cg(c,g^-1,h,h^-1,exp) + xi(c,g^-1,h^-1,k,exp)) mod exp;
                if phase <> 0 then
                    obs := true;
                    Print("Type 2 obstruction found from k=",k,"\n");
                    break;
                fi;
            od;
        fi;

        #### OBSTRUCTION 2
        # We are going to take the ugly route by bruteforce checking all the possibilities.
        # I am pretty sure it doesn't matter what generators we take, but I might want to verify this later

        if not obs then
            stab := Centralizer(group,p);
            if Size(stab)>1 then Print("Possible type 1 obstruction. Testing ",Size(stab)," elements.\n"); fi;
            for k in stab do
                phase := xi(c,g,h,k,exp) mod exp;
                if phase <> 0 then
                    obs := true;
                Print("Type 1 obstruction found from k=",k,"\n");
                    break;
                fi;
            od;
            if obs then
            else
                Print("Not obstructed\n");
                Add(out, p);
            fi;
        fi;
    od;
    return out;
end;

######################################################################
# List the gammas that are in the congruence subgroups for a given gamma
######################################################################


ListCongruenceSubgroup:=function(group, g, h)
    local ng, nh, a, b, c, d, list;

    list :=[];
    ng:=Order(g);
    nh:=Order(h);
    Print("elements are (mod ",ng,", mod ",ng,", mod ",nh,", mod ",nh,")\n");

    for a in [0..ng-1] do
        for b in [0..ng-1] do
            for c in [0..nh-1] do
                for d in [0..nh-1] do
                    if ((a *d -b * c) mod Gcd(ng,nh)=(1 mod Gcd(ng,nh))) and Gcd(a,b,ng)=1 and Gcd(c,d,nh)=1 and Gcd(a,c,ng,nh)=1 and Gcd(b,d,nh,ng)=1 and IsPairwiseConjugate(group, g, h, g^a*h^c,g^b*h^d) then
                        Print("contains (",a,",",b,",",c,",",d,")\n");
                        Add(list,[a,b,c,d]);
                    fi;
                od;
            od;
        od;
    od;
    return list;
end;
LCS:=ListCongruenceSubgroup;

ListCongruenceSubgroups:= function(group)
    local pairs, rpairs, p, g, h, ng, nh, a, b, c, d;
    pairs:=GetDistinctRankTwoAbelianSubgroups(group);
    rpairs:=Filtered(pairs,p->not IsCyclic(p));

    for p in rpairs do
        Print("\n",p,"\n");
        g := p.2;
        h := p.1;

        ListCongruenceSubgroup(group,g,h);
    od;
end;



getclassname:=function(g,tbl)
    local names, class, i;
    names := ClassNames( tbl );
    class := ConjugacyClasses(tbl);

    for i in [1..Size(class)] do
        if g in class[i] then
            return names[i];
        fi;
    od;
    return false;
end;


ZeroMinimum := function(list)
    if Length(list)=0 then return 0; else return Minimum(list); fi;
end;

ConstructCongruenceSubgroup:=function(x,y,z,ng,nh)
    local a,b,c,d,list;

    list :=[];

    for a in [0..ng-1] do
        for b in [0..ng-1] do
            for c in [0..nh-1] do
                for d in [0..nh-1] do
                    if ((a-1) mod y=0) and ((d-1) mod y=0) and (b mod z=0) and (c mod x=0) and ((a *d -b * c) mod Gcd(ng,nh)=(1 mod Gcd(ng,nh))) and Gcd(a,b,ng)=1 and Gcd(c,d,nh)=1 and Gcd(a,c,ng,nh)=1 and Gcd(b,d,nh,ng)=1 then
                        Add(list,[a,b,c,d]);
                    fi;
                od;
            od;
        od;
    od;
    return list;

end;

CongToString:=function(x,y,z)
    if x=1 and y=1 and z=1 then return "SL(2,Z)"; fi;
    if x=y and y=z then return JoinStringsWithSeparator(["Gamma(",PrintString(x),")"],""); fi;
    if y=1 and z=1 then return JoinStringsWithSeparator(["Gamma_0(",PrintString(x),")"],""); fi;
    if y=1 and x=1 then return JoinStringsWithSeparator(["Gamma^0(",PrintString(z),")"],""); fi;
    if x=y and z=1 then return JoinStringsWithSeparator(["Gamma_1(",PrintString(y),")"],""); fi;
    if x=1 and y=z then return JoinStringsWithSeparator(["Gamma^1(",PrintString(y),")"],""); fi;
    return JoinStringsWithSeparator(["Gamma(",PrintString(x),", ",PrintString(y),", ",PrintString(z),")"],"");
end;

# NameCongruenceSubgroup returns a name for the congruence subgroup if it is of the form
# Gamma_0(x)\cap Gamma(y)\cap Gamma^0(z)
NameCongruenceSubgroup:=function(ng,nh,list)
    local amin,bmin,cmin,dmin,x,y,z,comparelist;
    if nh=1 and ng=1 then return [1,1,1]; fi;

    amin := ZeroMinimum(Filtered(List(list,l->(l[1]-1) mod ng),l->l>0));
    bmin := ZeroMinimum(Filtered(List(list,l->l[2]),l->l>0));
    dmin := ZeroMinimum(Filtered(List(list,l->(l[4]-1) mod ng),l->l>0));
    cmin := ZeroMinimum(Filtered(List(list,l->l[3]),l->l>0));

    y := Gcd(Gcd(amin,dmin),ng,nh);
    x := Gcd(cmin,nh);
    z := Gcd(bmin,ng);
    Print(amin,bmin,cmin,dmin,"\n");
    Print(x,y,z,"\n");

    comparelist := ConstructCongruenceSubgroup(x,y,z,ng,nh);
    Print(comparelist,"\n");
    if IsEqualSet(comparelist,list) then
        return [x,y,z];
    else
        Print("Failed to determine subgroup\n");
        return [false,false,false];
    fi;
end;


######################################################################
# Functions for displaying the projective character tables of the centralizer of g in G
######################################################################

getclassname:=function(g,tbl)
    local names, class, i;
    names := ClassNames( tbl );
    class := ConjugacyClasses(tbl);

    for i in [1..Size(class)] do
        if g in class[i] then
            return names[i];
        fi;
    od;
    return false;
end;

ConstructShurCoveringHomomorphism:=function(G,g)
    local h,cent,cover,char,homs,covercent,coverchar,listofnames;
    cent:=Centralizer(G,g);
    char := CharacterTable(cent);
    Display(char);

    cover:=SchurCover(cent);

    if Size(cover)=Size(cent) then
        Print("Schur cover is isomorphic to target group.  No non-trivial projective representations.");
        return false;
    fi;

    coverchar:=CharacterTable(cover);
    Display(coverchar);


    homs := AllHomomorphismClasses(cover,cent);
    #Print("homs ",homs,"\n\n");
    covercent := Center(cover);
    homs := Filtered(homs,h->IsSurjective(h));
    #Print(homs,"\n\n");
    homs := Filtered(homs,h->IsSubgroup(covercent,Kernel(h)));
    #Print(homs,"\n\n");
    Print("Up to conjugation, there are ",Size(homs)," homomorphisms \n");

    h := homs[1];   # These are probably all the same up to isomorphism


    listofnames := List(ConjugacyClasses(coverchar),c->getclassname(Image(h,Representative(c)),char));
    Print(ClassNames(coverchar),"\n");
    Print(listofnames);
    return h;
end;

######################################################################
# Functions for explicit evaluations in terms of SL(2,Z)
# Most of these functions take an element of SL(2,Z) in terms of the string generated
# by Decompose([[a,b],[c,d]]).
######################################################################

# Code to decompose into generators of SL(2,Z).  This could be implemented much better,
# but I'm lazy.  This returns a string of the form "SQSST"
Decompose:=function(matrix)
    local a,b,c,d,an,bn,cn,dn,out;
    a:=matrix[1][1];
    b:=matrix[1][2];
    c:=matrix[2][1];
    d:=matrix[2][2];
    #Print(matrix,"\n");
    out := "";

    if d<>0 then
        if AbsInt(c)>AbsInt(d) or c=0 then
            an:=-b; bn:=a; cn:=-d; dn:=c;
            a:=an; b:=bn; c:=cn; d:=dn;
            out := Concatenation("S",out);
        fi;
        if c<0 then
            an:=-a; bn:=-b; cn:=-c; dn:=-d;
            a:=an; b:=bn; c:=cn; d:=dn;
            out := Concatenation("SS",out);
        fi;
        if d>0 then
            an:=a; bn:=b-a; cn:=c; dn:=d-c;
            a:=an; b:=bn; c:=cn; d:=dn;
            return Concatenation(Decompose([[a,b],[c,d]]),"T",out);
        fi;
        if d<0 then
            an:=a; bn:=b+a; cn:=c; dn:=d+c;
            a:=an; b:=bn; c:=cn; d:=dn;
            return Concatenation(Decompose([[a,b],[c,d]]),"Q",out);
        fi;
    fi;
    # c=0 now
    if c<0 then
        an:=-a; bn:=-b; cn:=-c; dn:=-d;
        a:=an; b:=bn; c:=cn; d:=dn;
        out := Concatenation("SS",out);
    fi;
    if a=0 then
        return Concatenation("S",out);
    fi;
    if a>0 then
        return Concatenation("T",Decompose([[a-1,-1],[1,0]]),out);
    fi;
    if a<0 then
        return Concatenation("Q",Decompose([[a+1,-1],[1,0]]),out);
    fi;
end;

# Calculates chi(g,h) under transformation by string, which is of the form "STTTSQSTST"
# for S = (0,-1,1,0), T = (1,1,0,1) and Q = T^-1.
# I think that doing this recursively is easiest.  The recursion comes from equation (46).
MultiplierPhase:=function(c,g,h,string,exp)
    local s2;

    if Size(string)=0 then
        return 0;
    fi;

    s2 := string{[2..Size(string)]};

    if string[1]='S' then
        return (-cg(c,h,g,g^-1,exp) + MultiplierPhase(c,h,g^-1,s2,exp)) mod exp;
    fi;

    if string[1]='T' then
        return (cg(c,g,g,h,exp) + MultiplierPhase(c,g,g*h,s2,exp)) mod exp;
    fi;

    if string[1]='Q' then
        return (-cg(c,g,g,(g^-1) * h,exp) + MultiplierPhase(c,g,(g^-1)*h,s2,exp)) mod exp;
    fi;

    return false;
end;

# returns (g,h)_gamma.  We are going to verify that these are conjugate to the initial g and h and find the correction to
# the multiplier phase
ghgamma:=function(g,h,string)
    local s2;

    if Size(string)=0 then
        return [g,h];
    fi;

    s2 := string{[2..Size(string)]};

    if string[1]='S' then
        return ghgamma(h,g^-1,s2);
    fi;

    if string[1]='T' then
        return ghgamma(g,g*h,s2);
    fi;

    if string[1]='Q' then
        return ghgamma(g,(g^-1)*h,s2);
    fi;

    return false;
end;

# Find the correction to the multiplier phase when (g,h)_gamma_k = (g,h)
MultiplierCorrection:=function(group,c,g1,h1,string,exp)
    local list,k,h2,g2;

    # find (g_h)_gamma
    list := ghgamma(g1,h1,string);
    g2:=list[1]; h2:=list[2];

    Print("{g,h}: ",g1," and ",h1,"\n");
    Print("gamma {g,h}:",g2," and ",h2,"\n");

    if not IsPairwiseConjugate(group,g1,h1,g2,h2) then
        return false;
    fi;

    k:=PairwiseConjugateRepresentative(group,g2,h2,g1,h1);

    return xi(c,g2,h2,k,exp) mod exp;
end;

#example: TotalMultiplierPhase(gg7,rr7,cc7,(),Random(gg7),[[1,1],[2,3]]);
TotalMultiplierPhase := function(group, R, cc, g, h, gamma)
    local c, string, p1, p2, exp;
    c := VectorToCocycle(cc,R);
    string := Decompose(gamma);

    exp:=Lcm(Homology(TensorWithIntegers(R),3));
    p1 := MultiplierPhase(c,g,h,string,exp);
    p2 := MultiplierCorrection(group,c,g,h,string,exp);

    Print ("phi_{g,h}(gamma tau,gamma z) = e(",p1/exp,")phi_gamma{g,h}(tau,z)");
    if p2 = false then
        Print(", which is not conjugate to {g,h}\n");
        return false;
    else
        Print(" = e(",(p1+p2 mod exp)/exp,")phi_{g,h}(tau,z)\n");
        return (p1+p2 mod exp)/exp;

    fi;
end;

TotalMultiplierPhase2 := function(group,c,g,h,gamma,exp)
    local string, p1, p2;
    string := Decompose(gamma);
    p1 := MultiplierPhase(c,g,h,string,exp);
    p2 := MultiplierCorrection(group,c,g,h,string,exp);

    Print ("phi_{g,h}(gamma tau,gamma z) = e(",p1/exp,")phi_gamma{g,h}(tau,z)");
    if p2 = false then
        Print(", which is not conjugate to {g,h}\n");
        return false;
    else
        Print(" = e(",(p1+p2 mod exp)/exp,")phi_{g,h}(tau,z)\n");
        return (p1+p2 mod exp)/exp;

    fi;
end;

# Returns a list with elements of the form [a,b,c,d,phase]
MultiplierPhaseForCongruenceSubgroup:=function(group, R, cc, g, h, cmax)
    local ng,nh,subgroup,list,a,b,c,d,gcd,elementmod,phase,k,dlist;
    ng := Order(g);
    nh := Order(h);
    subgroup := ListCongruenceSubgroup(group,g,h);
    #Print(subgroup);

    list := [];
    c:=0;
    while true do
        c:=c+1;

        dlist := Filtered([0..ng*c],p->Gcd(c,p)=1);
        #Print("\n",c,": dlist",dlist,"\n");
        for d in dlist do
            gcd := Gcdex(d,-c);
            a := gcd.coeff1;
            b := gcd.coeff2;

            for k in [0..ng-1] do
                elementmod := [(a+k*c) mod ng,(b+k*d) mod ng, c mod nh, d mod nh];
                #Print("\n",[a,b,c,d]," ",elementmod);
                if elementmod in subgroup then
                    #Print("is in subgroup");
                    phase := TotalMultiplierPhase(group,R,cc,g,h,[[a+k*c,b+k*d],[c,d]]);
                    Add(list,[a+k*c,b+k*d,c,d,phase]);
                    if Length(list)>=cmax then return list; fi;
                fi;
            od;
        od;
    od;
end;

# Finds the corresponding matrix
StringToMatrix:=function(string)
    local s2;

    if Size(string)=0 then
        return [[1,0],[0,1]];
    fi;

    s2 := string{[2..Size(string)]};

    if string[1]='S' then
        return [[0,-1],[1,0]]*StringToMatrix(s2);
    fi;

    if string[1]='T' then
        return [[1,1],[0,1]]*StringToMatrix(s2);
    fi;

    if string[1]='Q' then
        return [[1,-1],[0,1]]*StringToMatrix(s2);
    fi;

    return false;
end;

# Takes a list in the same format that MultiplierPhaseForCongruenceSubgroup outputs.
# Outputs an n such that phase=e(ncd).
EstimateMultiplierPhaseExponent:=function(phaselist)
    local phases,cds,denoms,delta,i,works,j;
    phases := List(phaselist,l->l[5]);
    cds := List(phaselist,l->l[3]*l[4]);
    denoms := List(phases,p->DenominatorRat(p)*Gcd(cds));
    delta := (Lcm(denoms)*Gcd(cds));

    #Print(phases,"\n",cds,"\n",delta);

    for i in [0..delta] do
        works := true;
        for j in [1..Length(cds)] do
            if ((cds[j]*i) mod delta) <> (phases[j]*delta) then works:=false; fi;
        od;
        if works then return (i/delta); fi;
    od;
    return false;
end;

# example: MultiplierPhaseForCongruenceSubgroupsForGroup(gg124,rr124,cc124,10,"~/Dropbox/GAP/multiplierforgg124.txt");
MultiplierPhaseForCongruenceSubgroupsForGroup:=function(group,R,cc,cmax,filename)
    local greps,centralizer,ccc,ccent,hreps,g,h,out1,out2,name,phasemult;
    ccc := ConjugacyClasses(group);
    PrintTo(filename);

    greps := List(ccc,c->Representative(c));
    for g in greps do

        AppendTo(filename,"\n\n\n\n# Multipliers for g=",g,"\n");

        centralizer := Centralizer(group,g);
        ccent := ConjugacyClasses(centralizer);
        hreps := List(ccent,c->Representative(c));

        for h in hreps do
            Print(g,"\n",h,"\n");
            out1 := ListCongruenceSubgroup(group,g,h);
            AppendTo(filename,"\n\n# g=",g,", h=",h);

            # Attempt to find a name for the congruence subgroup
            name := NameCongruenceSubgroup(Order(g),Order(h),out1);
            if name[1]=false then
                AppendTo(filename,"\nCongruence subgroup (mod ",Order(g),", mod ",Order(g),",mod ",Order(h),",mod ",Order(h),")");
                AppendTo(filename,"\n",out1);
            else
                AppendTo(filename,"\nCongruence subgroup is ",CongToString(name[1],name[2],name[3]));
            fi;

            # Attempt to find the multiplier phase
            out2 := MultiplierPhaseForCongruenceSubgroup(group, R, cc, g, h, cmax);

            phasemult := EstimateMultiplierPhaseExponent(out2);
            Print("Evaluated multiplier phase");

            if phasemult=false then
                AppendTo(filename,"\nSelected Multiplier Phases:\n",out2);
            else
                AppendTo(filename,"\n phi = e(",phasemult," cd)");
            fi;
        od;

    od;
end;

MultiplierPhaseFromMatrix:=function(c,g,h,gamma,exp)
  local string;
  string := Decompose(gamma);
  return MultiplierPhase(c,g,h,string,exp);
end;
